// src/backend/scheduler.cc - Copyright 2005, 2006, University
//               of Padova, dept. of Pure and Applied
//               Mathematics
//
// This file is part of SGPEMv2.
//
// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// SGPEMv2 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SGPEMv2; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

#include "concrete_environment.hh"
#include "concrete_history.hh"
#include "cpu_policy.hh"
#include "scheduler.hh"
#include "user_interrupt_exception.hh"

// Do not include full template definition in the header file
#include "singleton.tcc"

#include <glibmm/thread.h>
#include <iostream>
#include <cassert>
#include <memory>

using namespace std;
using namespace sgpem;

// Explicit template instantiation to allow to export symbols from the DSO.
template class SG_DLLEXPORT Singleton<Scheduler>;

typedef std::vector<DynamicProcess*>    Processes;
typedef std::vector<DynamicRequest*>    Requests;
typedef std::vector<DynamicSubRequest*> SubRequests;
typedef std::vector<DynamicThread*>     Threads;


Scheduler::Scheduler()
	: _ready_queue(NULL), _policy(NULL), _step_mutex()
{
}


ReadyQueue*
Scheduler::get_ready_queue()
{
  return _ready_queue;
}


CPUPolicy*
Scheduler::get_policy()
{
	return _policy;
}


//.............................................................................
//.............................................................................
//                           StepForward Preludium
//.............................................................................
//.............................................................................


static bool
validate_environment(ConcreteEnvironment& e)
{
	return true;
}



	// Introduces newly arrived processes.
	// Postcondition:
	// for each process p in next_snapshot
	// (
	// 	for each thread t belonging to p
	// 	(
	// 		if the arrival time of p is equal to front, then
	// 		(
	// 			the state of p is state_ready and
	// 			if the remaining time of t is zero, then
	//				the state of t is state_terminated
	// 		)
	// 	)
	// )
static void
update_future_processes(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if(dp.get_arrival_time() == front)
		{
			assert(dp.get_elapsed_time() == 0);
			typedef std::vector<DynamicThread*> DynamicThreads;
			DynamicThreads dts = dp.get_dynamic_threads();
			for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
 	   	{
				DynamicThread& dt = **it_dts;
				if(dt.get_arrival_time() == dp.get_elapsed_time())
				{
					dt.set_state(Schedulable::state_ready);
					// in this way we will never have threads ready having remaining time == 0
					if (dt.get_elapsed_time() == dt.get_total_cpu_time())
						dt.set_state(Schedulable::state_terminated);
				}
			}
    }
	}
}

	// Updates running process and thread
	// Postcondition:
	// for each thread t in next_snapshot
	// (
	// 	if the state of t is state_running then
	// 	(
	// 		the remaining time is one less than the one in the previous snapshot
	// 		if the remaining time is zero then
	// 			the state of t is state_terminated
	// 	)
	// )
static void
advance_running_process_and_thread(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if(dp.get_state() == Schedulable::state_running)
		{
			front = dp.get_elapsed_time();
			typedef std::vector<DynamicThread*> DynamicThreads;
			DynamicThreads dts = dp.get_dynamic_threads();
			for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
 	   	{
				DynamicThread& dt = **it_dts;
				if(dt.get_state() == Schedulable::state_running)
				{
					dt.decrease_remaining_time();
					if(dt.get_total_cpu_time() == dt.get_elapsed_time())
					{
						dt.set_state(Schedulable::state_terminated);
						typedef std::vector<DynamicRequest*> DynamicRequests;
						DynamicRequests& drs = dt.get_dynamic_requests();
						for(DynamicRequests::const_iterator it_drs = drs.begin(); it_drs != drs.end(); it_drs++)
						{
							DynamicRequest& dr = **it_drs;
							if(dr.get_state() != Request::state_exhausted)
							{
								typedef std::vector<DynamicSubRequest*> DynamicSubRequests;
								DynamicSubRequests dsrs = dr.get_dynamic_subrequests();
								for(DynamicSubRequests::const_iterator it_dsrs = dsrs.begin(); it_dsrs != dsrs.end(); it_dsrs++)
					 	   	{
									DynamicSubRequest& dsr = **it_dsrs;
									if(dsr.get_state() == Request::state_allocated)
									{
										while(dsr.get_remaining_time() != 0)
											dsr.decrease_remaining_time();
										/// Remove the subrequest (pointer) from the queue.
										bool removed = false;
										typedef Environment::SubRequestQueue SubRequestQueue;
										SubRequestQueue& queue = next_snapshot->get_request_queue(dsr.get_resource_key());
										for (SubRequestQueue::iterator it = queue.begin(); !removed && it != queue.end(); it++)
										{
											if ((*it) == &dsr)
											{
												queue.erase(it);
												removed = true;
											}
										}
									}
									dsr.set_state(Request::state_exhausted);
								}
							}
						}
					}
				}
			}
    }
	}
}

	// Introduces newly arrived				for(unsigned int i = 0; i < queue.size(); i++)
	// Postcondition:
	// for each process p in next_snapshot
	// (
	// 	for each thread t belonging to p
	// 	(
	// 		if the arrival time of p is equal to the elapsed time of p, then
	// 		(
	// 			the state of p is state_ready and
	// 			if the remaining time of t is zero, then
	//				the state of t is state_terminated
	// 		)
	// 	)
	// )
static void
update_future_threads(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if(dp.get_arrival_time() <= front)
		{
			front = dp.get_elapsed_time();
			typedef std::vector<DynamicThread*> DynamicThreads;
			DynamicThreads dts = dp.get_dynamic_threads();
			for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
 	   	{
				DynamicThread& dt = **it_dts;
				if(dt.get_arrival_time() == front)
				{
					dt.set_state(Schedulable::state_ready);
					// in this way we will never have threads ready having remaining time == 0
					if (dt.get_elapsed_time() == dt.get_total_cpu_time())
						dt.set_state(Schedulable::state_terminated);
				}
			}
    }
	}
}

	// Updates allocated resources and subresources
	// Postcondition:
	// for each subrequest sr in next_snapshot
	// (
	// 	if the state of sr is state_running then
	// 	(
	// 		the remaining time is one less than the one in the previous snapshot
	// 		if the remaining time is zero then
	// 			the state of sr is state_exhausted
	// 	)
	// )
static void
advance_allocated_requests_and_subrequests(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if(dp.get_state() == Schedulable::state_running)
		{
			typedef std::vector<DynamicThread*> DynamicThreads;
			DynamicThreads dts = dp.get_dynamic_threads();
			for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
 	   	{
				DynamicThread& dt = **it_dts;
				if(dt.get_state() == Schedulable::state_running)
				{
					typedef std::vector<DynamicRequest*> DynamicRequests;
					DynamicRequests& drs = dt.get_dynamic_requests();
					for(DynamicRequests::const_iterator it_drs = drs.begin(); it_drs != drs.end(); it_drs++)
					{
						DynamicRequest& dr = **it_drs;
						if(dr.get_state() == Request::state_allocated)
						{
							typedef std::vector<DynamicSubRequest*> DynamicSubRequests;
							DynamicSubRequests dsrs = dr.get_dynamic_subrequests();
							for(DynamicSubRequests::const_iterator it_dsrs = dsrs.begin(); it_dsrs != dsrs.end(); it_dsrs++)
				 	   	{
								DynamicSubRequest& dsr = **it_dsrs;
								if(dsr.get_state() == Request::state_allocated)
								{
									dsr.decrease_remaining_time();
									if(dsr.get_remaining_time() == 0)
									{
										dsr.set_state(Request::state_exhausted);
	        				  /// Remove the subrequest (pointer) from the queue.
					          bool removed = false;
										typedef Environment::SubRequestQueue SubRequestQueue;
										SubRequestQueue& queue = next_snapshot->get_request_queue(dsr.get_resource_key());
				    	      for (SubRequestQueue::iterator it = queue.begin(); !removed && it != queue.end(); it++)
										{
				        	    if ((*it) == &dsr)
        				    	{
        				      	queue.erase(it);
				          	    removed = true;
				            	}
										}
				        	}
								}
							}
						}
					}
				}
			}
    }
	}
}

	// Updates unallocated subrequests
	// Postcondition:
	// for each subrequest sr in next_snapshot
	// (
	// 	let r be the resource specified in the subrequest
	// 	let q be the queue associated with r
	// 	let s be the state of sr
	// 	let p be the number of places of r
	// 	if sr is in q then
	// 	if the position of sr in q is lesser or equal to p then
	// 		the state of sr is state_allocable
	// 	else
	// 		the state of sr is state_unallocable
	// )
	// for each thread t in next_snapshot
	// (
	// 	let a be the number of requests of t whose state is state_allocable
	// 	let u be the number of requests of t whose state is state_unallocable
	//	if a is not zero and u is zero then
	// 		the state of t is state_ready
	// 	if a is not zero and u is not zero then
	//		the state of t is state_blocked
	// )
static void
update_nonrunning_threads_and_unallocated_subrequests(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		typedef std::vector<DynamicThread*> DynamicThreads;
		DynamicThreads dts = dp.get_dynamic_threads();
		for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
	   	{
			DynamicThread& dt = **it_dts;
			if(dt.get_state() == Schedulable::state_ready || dt.get_state() == Schedulable::state_blocked)
			{
				bool blocked = false;
				typedef std::vector<DynamicRequest*> DynamicRequests;
				DynamicRequests drs = dt.get_dynamic_requests();
				for(DynamicRequests::const_iterator it_drs = drs.begin(); it_drs != drs.end(); it_drs++)
				{
					DynamicRequest& dr = **it_drs;
					if(dr.get_state() == Request::state_allocable || dr.get_state() == Request::state_unallocable)
					{
						typedef std::vector<DynamicSubRequest*> DynamicSubRequests;
						DynamicSubRequests dsrs = dr.get_dynamic_subrequests();
						for(DynamicSubRequests::const_iterator it_dsrs = dsrs.begin(); it_dsrs != dsrs.end(); it_dsrs++)
			 	   	{
							DynamicSubRequest& dsr = **it_dsrs;
							assert(dsr.get_state() == Request::state_allocable || dsr.get_state() == Request::state_unallocable);
			        unsigned int position = 0;
							Environment::SubRequestQueue& queue = next_snapshot->get_request_queue(dsr.get_resource_key());
							Environment::SubRequestQueue::iterator it = queue.begin();
			        while (it != queue.end())
       				{
			          if (*it == &dsr)
       				    break;
								it++;
								position++;
			        }
			        /// Watch out: in a resource with 2 places, 0 and 1 are valid queue
			        /// positions, 2 is right one place out.
							if (position >= next_snapshot->get_resources().find(dsr.get_resource_key())->second->get_places())
								dsr.set_state(Request::state_unallocable);
							else
			        	dsr.set_state(Request::state_allocable);
						}
					}
					if (dr.get_state() == Request::state_unallocable)
						blocked = true;
				}
				if (blocked)
					dt.set_state(Schedulable::state_blocked);
				else
					dt.set_state(Schedulable::state_ready);
			}
		}
	}
}

	// prepares the environment advancing counters and updating states appropriately
static void
advance_and_update_all(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	// Checks if there are future processes that should be marked as ready, 
	// since their arrival time is equal to the simulation current instant.
	update_future_processes(front, next_snapshot);

	// Updates the counters of the running entities.
	// Checks if the running thread (if any) should be marked as 
	// terminated.
	advance_running_process_and_thread(front, next_snapshot);

	// Checks if there are subrequests associated to the running thread that have 
	// just exausted, or if the old running thread has terminated. Then adjust 
	// the subrequests' list accordingly (freeing associated resources);
	advance_allocated_requests_and_subrequests(front, next_snapshot);

	// Checks if there are future threads that should be marked as ready, 
	// since their arrival time is equal to their owner process' elapsed time.
	update_future_threads(front, next_snapshot);

	// If any thread was blocked on a now freed resource, pass it from blocked 
	// to ready state, by discovering just-now-allocable requests.
	// If any thread was ready on a now busy resource, pass it from ready to 
	// blocked state, by discovering just-now-unallocable requests.
	// This is obtained by 
	update_nonrunning_threads_and_unallocated_subrequests(front, next_snapshot);

}


	// Function ?? FIXME
	// Postcondition:
	// if the policy is priority-preemptable or the quantum is finished then
	// 	for each thread t in next_snapshot
	// 	(
	// 		if the policy is priority-preemptable and the state of t was state_running
	// 			the state of t is state_ready
	// 		if the policy is quantum-preemptable and the state of t was state_running and the quantum has finished
	// 			the state of t is state_ready
	// 	)
static void
manage_preemption(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot, bool preemptible_policy, unsigned int time_slice)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		typedef std::vector<DynamicThread*> DynamicThreads;
		DynamicThreads dts = dp.get_dynamic_threads();
		for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
		{
			DynamicThread& dt = **it_dts;
			if(dt.get_state() == Schedulable::state_running)
			{
				if(preemptible_policy || time_slice == front - dt.get_last_acquisition())
				{
					dt.set_state(Schedulable::state_ready);
					dt.set_last_release(front);
				}
			}
		}
	}
}



	// FIXME: document me please
static void
build_ready_queue(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	ReadyQueue& queue = next_snapshot->get_sorted_queue();
	assert(queue.size() == 0);

	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		typedef std::vector<DynamicThread*> DynamicThreads;
		DynamicThreads dts = dp.get_dynamic_threads();
		for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
		{
			DynamicThread& dt = **it_dts;
			if(dt.get_state() == Schedulable::state_ready)
			{
				queue.append(dt);
			}
		}
	}
}


	// returns true if and only if the snapshot contains a candidate running thread,
	// the thread's state is state_running.
static bool
find_a_candidate(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	// look for a currently running process first
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if (dp.get_state() == Schedulable::state_running)
		{
			typedef std::vector<DynamicThread*> DynamicThreads;
			DynamicThreads dts = dp.get_dynamic_threads();
			for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
			{
				DynamicThread& dt = **it_dts;
				if(dt.get_state() == Schedulable::state_running)
					return true;
			}
		}
	}
	// if no process has been found, select the first on the ready queue.
	ReadyQueue& queue = next_snapshot->get_sorted_queue();
	if (queue.size() == 0)
		return false;
	DynamicThread& candidate = dynamic_cast<DynamicThread&>(queue.get_item_at(0));
	candidate.set_state(Schedulable::state_running);
	// HACK HACK HACK
	// we do not remove the candidate from the ready queue. this information is useful
	// since we chose to set the last_aquisition and the last_release just after a
	// successful selection. see try_to_run().
	return true;
}







	// returns true if and oly if the simulation is terminated
static bool
check_if_simulation_is_terminated(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if (dp.get_state() == Schedulable::state_running)
			return false;
		if (dp.get_state() == Schedulable::state_future)
			return false;
	}
	return true;
}


static void
print_queue(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	ReadyQueue& queue = next_snapshot->get_sorted_queue();
	for(unsigned int i = 0; i < queue.size(); i++)
	{
		std::cout << queue.get_item_at(i).get_name();
	}
}






static bool
try_to_run(unsigned int front, auto_ptr<ConcreteEnvironment>& next_snapshot)
{
	bool success = false;
	typedef std::vector<Process*> Processes;
	Processes ps = next_snapshot->get_processes();
	for(Processes::const_iterator it_ps = ps.begin(); it_ps != ps.end(); it_ps++)
	{
		DynamicProcess& dp = dynamic_cast<DynamicProcess&>(**it_ps);
		if (dp.get_state() == Schedulable::state_running)
		{
			typedef std::vector<DynamicThread*> DynamicThreads;
			DynamicThreads dts = dp.get_dynamic_threads();
			for(DynamicThreads::const_iterator it_dts = dts.begin(); it_dts != dts.end(); it_dts++)
			{
				DynamicThread& dt = **it_dts;
				if(dt.get_state() == Schedulable::state_running)
				{
					// this is our candidate
					success = true;
					// let's see if it is runnable or if it blocks:
					typedef std::vector<DynamicRequest*> DynamicRequests;
					DynamicRequests drs = dt.get_dynamic_requests();
					for(DynamicRequests::const_iterator it_drs = drs.begin(); it_drs != drs.end(); it_drs++)
					{
						DynamicRequest& dr = **it_drs;
						// if it's time to do it, raise a request
						if(dr.get_state() == Request::state_future && dr.get_instant() == dt.get_elapsed_time())
						{
							typedef std::vector<DynamicSubRequest*> DynamicSubRequests;
							DynamicSubRequests dsrs = dr.get_dynamic_subrequests();
							for(DynamicSubRequests::const_iterator it_dsrs = dsrs.begin(); it_dsrs != dsrs.end(); it_dsrs++)
			 	   		{
								DynamicSubRequest& dsr = **it_dsrs;
				        assert(dsr.get_state() == Request::state_future);
								Environment::SubRequestQueue& queue = next_snapshot->get_request_queue(dsr.get_resource_key());
								/// Enqueue the subrequest at the back of the queue.
								queue.push_back(&dsr);

				        /// TODO: right here, right now we should call the resource policy to
				        /// update the queue. Updates the state of the subrequest depending
				        /// on the position in the queue, as explained before.
								unsigned int places = next_snapshot->get_resources().find(dsr.get_resource_key())->second->get_places();
				        dsr.set_state(queue.size() >  places ? Request::state_unallocable : Request::state_allocable);
							} // for each subrequest
							// if it is actually allocable, allocate it
			        if (dr.get_state() == Request::state_allocable)
  	  		    {
								for(DynamicSubRequests::const_iterator it_dsrs = dsrs.begin(); it_dsrs != dsrs.end(); it_dsrs++)
				 	   		{
									DynamicSubRequest& dsr = **it_dsrs;
		      	  		assert(dsr.get_state() == Request::state_allocable);
		        	  	dsr.set_state(Request::state_allocated);
								}
							}
						} // end of request raising
						// if it does exist at least one unallocable request, the thread may not run!
		        if (dr.get_state() == Request::state_unallocable)
    		    {
							dt.set_state(Schedulable::state_blocked);
							success = false;
						}
					} // end for loop over all running thread requests

					// HACK HACK HACK
					// this check is legal, since all pointers in the sorted queue should be valid.
					// if the current thread is the first in the ready queue, we need to remove it
					if (next_snapshot->get_sorted_queue().size() != 0 &&
						&next_snapshot->get_sorted_queue().get_item_at(0) == &dt)
					{
						// in case of success in allocating the cpu, we must update the last_aquisition
						if (success)
							dt.set_last_acquisition(front);
						next_snapshot->get_sorted_queue().erase_first();
					} // if the current thread was the one that was previously running
					else
					{
						// we must update the last_release
						if (!success)
							dt.set_last_release(front);
						else
							return success;
						//cpu_policy.sort_queue();
	        }



				} // end things to do if thread was running
			} // end for loop over all threads of a running process
		} // end if process is running
	} // end for loop over all processes
	return false;
}// end main loop

//.............................................................................
//.............................................................................
//              The Return of the Son of the Monster Step Forward
//
//     The refactoring of this method has been carried on listening to Frank 
//             Zappa's The Return of the Son of the Monster Magnet.
//                 The magic word for tonight is.. procedural!
//.............................................................................
//.............................................................................

bool
Scheduler::step_forward(ConcreteHistory& concrete_history, CPUPolicy& cpu_policy)
	throw(UserInterruptException)
{
	// Preconditions:
	assert (concrete_history.get_size() > 0);

	// Concurrency:
	// This very method should be exclusive: no concurrent behaviour, from when 
	// we store a readyqueue and policy pointer for the user-policy to retrieve, 
	// to when the policy returns.
  Glib::Mutex::Lock lock(_step_mutex);

	// The instant we are going to build a snapshot for:
	// if history has size 1, then the snapshot we will add will describe the
	// environment at instan 0, i.e. (1 - 1).
	unsigned int next_instant = concrete_history.get_size() - 1;

	// The snapshot we are building: it is built as a copy og the last one.
	// We use an auto_ptr since we've some exceptions in the coming...
	auto_ptr<ConcreteEnvironment> next_snapshot(new ConcreteEnvironment(concrete_history.get_last_environment()));

	// Temporarily set the _ready_queue param and the _policy one for 
	// use from external plugins
	_policy = &cpu_policy;
	_ready_queue = &next_snapshot->get_sorted_queue();

	// begin procedure step forward
	// ahw, I lowe soo munch all thissh olt fasshoned procedureal programing stylus
	advance_and_update_all(next_instant, next_snapshot);

	try
  {
		// If the user-policy is preemptible, or if the assigned time slice has 
		// terminated, pass the running thread to ready state.
		manage_preemption(next_instant, next_snapshot, cpu_policy.is_pre_emptive(), cpu_policy.get_time_slice());

		// Build the schedulables ready queue
		build_ready_queue(next_instant, next_snapshot);

		// and pass it to the policy itself
//		if (next_snapshot->get_sorted_queue().size() != 0)
//			cpu_policy.sort_queue();


		bool running = false;
		bool idle = false;
		bool terminated = false;

		do
		{
			if (find_a_candidate(next_instant, next_snapshot))
				running = try_to_run(next_instant, next_snapshot);
			else
				if (check_if_simulation_is_terminated(next_instant, next_snapshot))
					terminated = true;
				else
					idle = true;
		}
		while (!running && !idle && !terminated);

		// append the new snapshot, releasing the auto_ptr!
		concrete_history.append_new_environment(next_snapshot.release());

		// Reset values that the policy doesn't need anymore
		_policy      = NULL;
		_ready_queue = NULL;

		return !terminated; // watch out for the !
	}
	catch(UserInterruptException& e)
	{
		// Reset values that the policy doesn't need anymore
		_policy      = NULL;
		_ready_queue = NULL;

		//  Do we need to update/reset something else?
		// Going up unwinding the stack, tell:
		//    - the user that the policy sucks
		//    - SimulationController that everything stopped
		throw;
	}	
}
